c  Here is the program to find periodic orbits in  n-d KS system.
c  For laziness reasons I let naglib do the integration and inverting an
c  n+ncyc dim matrix. The essentials of the multi point method is in the
c  subroutine solve, but to change the program to do any other system it
c  should only be necessary to change the subroutine fcn which contains the
c  ODE plus bits and pieces of the main program (and all relevant dimensions
c  given as parameters). The input should be in guess.dat in the form
c  first line: ncyc
c  following lines: ((x(j,i),j=2,n),i=1,ncyc) with line shifts at least
c                   with every i.
c  repeat ad nauseam


      program kscyc
c  n is the dimension of the system, m is the maximal cycle length.
      parameter (n=16,m=10,nd=n*(n+1))

      real*8 xi(n,m),xf(n,m),xjac(n,n,m),xt(m),yi(n),yf(n),t
     & ,jac(n,n),xdum(nd),f(nd),v(n,m),vis,pvec(n),ev(n),evi(n),err

c  vis is the parameter of the system
      common /vis/ vis
      vis=0.02992

      t=0
c  reading initial guess from the file guess.dat. x(1)=0 is the Poincare sect.
      open(1,file='guess.dat',status='old')
 20   read(1,*,end=99) ncyc
      write(*,*) 'ncyc ',ncyc
      do i=1,ncyc
         read(1,*) (xi(j,i),j=2,n)
         xi(1,i)=0
      enddo
      nc=0
 10   nc=nc+1

c  calculating v=dx/dt, t=0, for use in subroutine solve
      do i=1,ncyc
         do j=1,n
            xdum(j)=xi(j,i)
         enddo
         call fcn(t,xdum,f)
         if(f(1).lt.0) goto 20
         do j=1,n
            v(j,i)=f(j)
         enddo
      enddo

c  pvec is a normalvector to the Poincare surface of section (if the surface
c  of section is not a hyperplane this needs to be modified severely)
      pvec(1)=1
      do i=2,n
         pvec(i)=0
      enddo

c  integration: xi is the initial point, xf=f(xi), xjac the correspoing
c  Jacobian, xt the time - all of this for each point on the surface
c  of section.
      do i=1,ncyc
         do j=1,n
            yi(j)=xi(j,i)
         enddo
         call inte(yi,yf,t,jac)
         xt(i)=t
         do j=1,n
            xf(j,i)=yf(j)
            do k=1,n
               xjac(j,k,i)=jac(j,k)
            enddo
         enddo
      enddo

c  solve dF/dx (x'-x)=-F(x), where Fi(x)=xi-f(x{i-1})
c  jac is the cumulative Jacobian and err is |x'-x|
      call solve(ncyc,xi,xf,xjac,v,pvec,jac,err)

      write(*,*) 'error ',err
      if(err.gt.5) goto 20
      if(err.gt.1.d-10.and.nc.lt.20) goto 10
      open(2,file='cyc.dat',status='unknown')
      write(2,*) ncyc
      write(2,*)
      do i=1,ncyc
         write(2,*) (xi(j,i),j=2,n)
         write(2,*)
      enddo
      t=0
      do i=1,ncyc
         t=t+xt(i)
      enddo
      write(2,*) t
      write(2,*)
      call eigen(jac,ev,evi)
      do i=1,n
         write(2,*) ev(i),evi(i)
      enddo
      write(2,*)
      goto 20
 99   stop
      end


c  subroutine to set up the naglib integration.
c  fcn gives the ODE. g determines the Poincare surface of section.
c  y(i), i=1,n, contains the coordinates; i=n+1,n*(n+1) contains the Jacobian.
      subroutine inte(yi,yf,t,jac)
      parameter (n=16,nd=n*(n+1))
      real*8 y(nd),tol,w(nd,7),ti,tf,hmax,
     & t,yi(n),yf(n),jac(n,n)

      external fcn,g

      do i=1,n
         y(i)=yi(i)
         do j=1,n
            y(n*i+j)=0
         enddo
         y((n+1)*i)=1
      enddo

      ifail=0
      tol=1.d-10
      irelab=0
      hmax=0
      ti=0

      do i=1,2
         tf=ti+1.d-4
         tol=abs(tol)
         call d02baf(ti,tf,nd,y,tol,fcn,w,ifail)
         tol=abs(tol)
         tf=ti+1000
         call d02bhf(ti,tf,nd,y,tol,irelab,hmax,fcn,g,w,ifail)
      enddo

      t=ti
      do i=1,n
         yf(i)=y(i)
         do j=1,n
            jac(i,j)=y(n*i+j)
         enddo
      enddo
      return
      end        


      subroutine fcn(t,y,f)
      parameter (n=16,nd=n*(n+1))
      real*8 f(nd),y(nd),vis,t,dv(n,n),jac(n,n),mat(n,n)
      common /vis/ vis

c  equations for the ODE
      do i=1,n
         f(i)=(i**2-vis*i**4)*y(i)
      enddo
      do i=1,n
         do j=i+1,n
            f(i)=f(i)+i*y(j)*y(j-i)
         enddo
         do j=1,i-1
            f(i)=f(i)-i*y(j)*y(i-j)
         enddo
         do j=i-n,-1
            f(i)=f(i)+i*y(-j)*y(i-j)
         enddo
      enddo
      
c  equations for the Jacobian. dv=df/dx, jac is the jacobian and mat=djac/dt
      do i=1,n
         do j=1,n
            mat(i,j)=0
            dv(i,j)=0
            jac(i,j)=y(n*i+j)
         enddo
      enddo

      do i=1,n
         dv(i,i)=i**2-vis*i**4
         do j=i+1,n
            dv(i,j)=dv(i,j)+i*y(j-i)
            dv(i,j-i)=dv(i,j-i)+i*y(j)
         enddo
         do j=1,i-1
            dv(i,j)=dv(i,j)-i*y(i-j)
            dv(i,i-j)=dv(i,i-j)-i*y(j)
         enddo
         do j=i-n,-1
            dv(i,-j)=dv(i,-j)+i*y(i-j)
            dv(i,i-j)=dv(i,i-j)+i*y(-j)
         enddo
      enddo

      do i=1,n
         do j=1,n
            do k=1,n
               mat(i,j)=mat(i,j)+dv(i,k)*jac(k,j)
            enddo
         enddo
      enddo

      do i=1,n
         do j=1,n
            f(n*i+j)=mat(i,j)
         enddo
      enddo
      return
      end
      
      
c  g=0 determines the Poincare surface of section. 
      real*8 function g(t,y)
      parameter (n=16,nd=n*(n+1))
      real*8 t,y(nd)
      g=y(1)
      return
      end


      subroutine solve(ncyc,xi,xf,xjac,v,pvec,jac,err)
      parameter (n=16,m=10,nd=n+m)
      real*8 xi(n,m),xf(n,m),xjac(n,n,m),v(n,m),pvec(n),cum(n,n,m)
     &  ,mat(nd,nd),vec(nd),augcol(n,m,m),f(n,m)
     &  , augf(m),matinv(nd,nd),jac(n,n),err

c  Augcol is an augmentation of the Newton method matrix with ncyc
c  columns that contains dxi/dt at the right place. The idea is that
c  we add small vectors along the time direction to the solution so that
c  it will be on the surface of section. augf will finally contain the 
c  size of the vectors.
c  Likewise we are also augmenting with ncyc rows that contains pvec - 
c  a vector normal to the surf. of sec.
      do i=1,n
         do j=1,ncyc
            do k=1,ncyc
               augcol(i,j,k)=0
            enddo
            augcol(i,j,j)=v(i,j)
         enddo
      enddo

c  cum is the cumulative Jacobian (starting i=ncyc)
      do i=1,n
         do j=1,n
            cum(i,j,1)=xjac(i,j,ncyc)
         enddo
      enddo
      do l=2,ncyc
         do i=1,n
            do j=1,n
               cum(i,j,l)=0
               do k=1,n
                  cum(i,j,l)=cum(i,j,l)+xjac(i,k,l-1)*cum(k,j,l-1)
               enddo
            enddo
         enddo
      enddo
      do i=1,n
         do j=1,n
            jac(i,j)=cum(i,j,ncyc)
         enddo
      enddo

c  f is -F(x)
      do i=1,ncyc
         i_1=mod(i-2+ncyc,ncyc)+1
         do j=1,n
            f(j,i)=xf(j,i_1)-xi(j,i)
         enddo
         augf(i)=0
      enddo
      
c  Here we are getting rid of the Jacobians below the diagonal in the
c  Newton method matrix - instead creating a column of the cumulative
c  Jacobians.
      do i=2,ncyc
         do j=1,i-1
            do k=1,n
               do l=1,n
                  augcol(k,i,j)=augcol(k,i,j)+
     &                 xjac(k,l,i-1)*augcol(l,i-1,j)
               enddo
            enddo
         enddo
         do k=1,n
            do l=1,n
               f(k,i)=f(k,i)+xjac(k,l,i-1)*f(l,i-1)
            enddo
         enddo
      enddo
      
c  At the lowest point of the diagonal we now have 1-jac
      do i=1,n
         do j=1,n
            mat(i,j)=-cum(i,j,ncyc)
         enddo
         mat(i,i)=mat(i,i)+1
      enddo
c  Augmenting 1-jac with ncyc columns
      do i=1,n
         do j=1,ncyc
            mat(i,n+j)=augcol(i,ncyc,j)
         enddo
      enddo
c  Augmenting 1-jac with ncyc rows
      do j=1,n
         do i=1,ncyc
            mat(n+i,j)=0
         enddo
         mat(n+ncyc,j)=pvec(j)
      enddo
      do i=1,ncyc
         do j=1,ncyc
            mat(n+i,n+j)=0
         enddo
      enddo
      do i=1,ncyc-1
         do j=1,n
            do k=1,n
               mat(n+i,k)=mat(n+i,k)+pvec(j)*cum(j,k,i)
            enddo
            do k=1,ncyc
               mat(n+i,n+k)=mat(n+i,n+k)-pvec(j)*augcol(j,i,k)
            enddo
            augf(i)=augf(i)-pvec(j)*f(j,i)
         enddo
      enddo

c  inverting the augmented 1-jac
      call inv(mat,matinv,ncyc)
      do i=1,n+ncyc
         vec(i)=0
         do j=1,n
            vec(i)=vec(i)+matinv(i,j)*f(j,ncyc)
         enddo
         do j=1,ncyc
            vec(i)=vec(i)+matinv(i,n+j)*augf(j)
         enddo
      enddo
      do i=1,n
         f(i,ncyc)=vec(i)
      enddo
      do i=1,ncyc
         augf(i)=vec(n+i)
      enddo

c  getting rid of the column of cumulative Jacobians
      do i=1,ncyc-1
         do j=1,n
            do k=1,n
               f(j,i)=f(j,i)+cum(j,k,i)*f(k,ncyc)
            enddo
         enddo
      enddo

c  getting rid of the augmenting columns
      do i=1,ncyc-1
         do j=1,n
            do k=1,ncyc
               f(j,i)=f(j,i)-augcol(j,i,k)*augf(k)
            enddo
         enddo
      enddo

c  finished
      err=0
      do i=1,ncyc
         do j=1,n
            err=err+f(j,i)**2
            xi(j,i)=xi(j,i)+f(j,i)
         enddo
      enddo
      err=sqrt(err)

      return
      end

         
c  naglib matrix inversion routine
      subroutine inv(mat,matinv,ncyc)
      parameter (n=16,m=10,nd=n+m)
      double precision mat(nd,nd),b(nd,nd),matinv(nd,nd),wkspce(nd),
     &     aa(nd,nd),bb(nd,nd)
      ifail=0
      ndim=n+ncyc
      do i=1,ndim
         do j=1,ndim
            b(i,j)=0
         enddo
         b(i,i)=1
      enddo
      call f04aef(mat,nd,b,nd,ndim,ndim,matinv,nd,
     &     wkspce,aa,nd,bb,nd,ifail)
      return
      end


c  naglib eigenvalue (and eigenvector) routine
      subroutine eigen(jac,ev,evi)
      parameter (n=16)
      double precision jac(n,n),ev(n),evi(n)
     & ,vr(n,n),vi(n,n)
      integer ifail,intger(n)
      ifail=0
      call f02agf(jac,n,n,ev,evi,vr,n,vi,n,intger,ifail)
      return
      end







         
